# sage_setup: distribution = sagemath-modules
"""
Equivalence Testing

AUTHORS:

- Anna Haensch (2014-12-01): added test for rational isometry
"""
from typing import Any

from sage.arith.misc import (hilbert_symbol,
                             GCD,
                             is_prime,
                             legendre_symbol,
                             prime_divisors,
                             valuation)
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ


##############################################################################
# Routines to test if two quadratic forms over ZZ are globally equivalent.   #
# (For now, we require both forms to be positive definite.)                  #
##############################################################################

def is_globally_equivalent_to(self, other, return_matrix=False) -> bool | Any:
    r"""
    Determine if the current quadratic form is equivalent to the
    given form over `\ZZ`.

    If ``return_matrix`` is True, then we return the transformation
    matrix `M` so that ``self(M) == other``.

    INPUT:

    - ``self``, ``other`` -- positive definite integral quadratic forms

    - ``return_matrix`` -- boolean (default: ``False``); return
      the transformation matrix instead of a boolean

    OUTPUT:

    - if ``return_matrix`` is ``False``: a boolean

    - if ``return_matrix`` is ``True``: either ``False`` or the
      transformation matrix

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: M = Matrix(ZZ, 4, 4, [1,2,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1])
        sage: Q1 = Q(M)
        sage: Q.is_globally_equivalent_to(Q1)                                           # needs sage.libs.pari
        True
        sage: MM = Q.is_globally_equivalent_to(Q1, return_matrix=True)                  # needs sage.libs.pari
        sage: Q(MM) == Q1                                                               # needs sage.libs.pari
        True

    ::

        sage: # needs sage.libs.pari
        sage: Q1 = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q2 = QuadraticForm(ZZ, 3, [2, 1, 2, 2, 1, 3])
        sage: Q3 = QuadraticForm(ZZ, 3, [8, 6, 5, 3, 4, 2])
        sage: Q1.is_globally_equivalent_to(Q2)
        False
        sage: Q1.is_globally_equivalent_to(Q2, return_matrix=True)
        False
        sage: Q1.is_globally_equivalent_to(Q3)
        True
        sage: M = Q1.is_globally_equivalent_to(Q3, True); M
        [-1 -1  0]
        [ 1  1  1]
        [-1  0  0]
        sage: Q1(M) == Q3
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1, -1])
        sage: Q.is_globally_equivalent_to(Q)                                            # needs sage.libs.pari
        Traceback (most recent call last):
        ...
        ValueError: not a definite form in QuadraticForm.is_globally_equivalent_to()

    ALGORITHM: this uses the PARI function :pari:`qfisom`, implementing
    an algorithm by Plesken and Souvignier.

    TESTS:

    :issue:`27749` is fixed::

        sage: Q = QuadraticForm(ZZ, 2, [2, 3, 5])
        sage: P = QuadraticForm(ZZ, 2, [8, 6, 5])
        sage: Q.is_globally_equivalent_to(P)                                            # needs sage.libs.pari
        False
        sage: P.is_globally_equivalent_to(Q)                                            # needs sage.libs.pari
        False
    """
    from sage.quadratic_forms.quadratic_form import QuadraticForm

    # Check that other is a QuadraticForm
    if not isinstance(other, QuadraticForm):
        raise TypeError("you must compare two quadratic forms, but the argument is not a quadratic form")

    # only for definite forms
    if not self.is_definite() or not other.is_definite():
        raise ValueError("not a definite form in QuadraticForm.is_globally_equivalent_to()")

    mat = other.__pari__().qfisom(self)
    if not mat:
        return False

    if return_matrix:
        return mat.sage()
    else:
        return True


def is_locally_equivalent_to(self, other, check_primes_only=False,
                             force_jordan_equivalence_test=False) -> bool:
    r"""
    Determine if the current quadratic form (defined over `\ZZ`) is
    locally equivalent to the given form over the real numbers and the
    `p`-adic integers for every prime `p`.

    This works by comparing the local Jordan decompositions at every
    prime, and the dimension and signature at the real place.

    INPUT:

    - ``other`` -- a :class:`QuadraticForm`

    OUTPUT: boolean

    EXAMPLES::

        sage: Q1 = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q2 = QuadraticForm(ZZ, 3, [2, 1, 2, 2, 1, 3])
        sage: Q1.is_globally_equivalent_to(Q2)                                          # needs sage.libs.pari
        False
        sage: Q1.is_locally_equivalent_to(Q2)                                           # needs sage.libs.pari
        True
    """
    # TO IMPLEMENT:
    if self.det() == 0:
        raise NotImplementedError("OOps!  We need to think about whether this still works for degenerate forms...  especially check the signature.")

    # Check that both forms have the same dimension and base ring
    if (self.dim() != other.dim()) or (self.base_ring() != other.base_ring()):
        return False

    # Check that the determinant and level agree
    if (self.det() != other.det()) or (self.level() != other.level()):
        return False

    # -----------------------------------------------------

    # Test equivalence over the real numbers
    if self.signature() != other.signature():
        return False

    # Test equivalence over Z_p for all primes
    if (self.base_ring() == ZZ) and (not force_jordan_equivalence_test):

        # Test equivalence with Conway-Sloane genus symbols (default over ZZ)
        if self.CS_genus_symbol_list() != other.CS_genus_symbol_list():
            return False
    else:
        # Test equivalence via the O'Meara criterion.
        for p in prime_divisors(ZZ(2) * self.det()):
            if not self.has_equivalent_Jordan_decomposition_at_prime(other, p):
                return False

    # All tests have passed!
    return True


def has_equivalent_Jordan_decomposition_at_prime(self, other, p) -> bool:
    """
    Determine if the given quadratic form has a Jordan decomposition
    equivalent to that of ``self``.

    INPUT:

    - ``other`` -- a :class:`QuadraticForm`

    OUTPUT: boolean

    EXAMPLES::

        sage: Q1 = QuadraticForm(ZZ, 3, [1, 0, -1, 1, 0, 3])
        sage: Q2 = QuadraticForm(ZZ, 3, [1, 0, 0, 2, -2, 6])
        sage: Q3 = QuadraticForm(ZZ, 3, [1, 0, 0, 1, 0, 11])
        sage: [Q1.level(), Q2.level(), Q3.level()]
        [44, 44, 44]

        sage: # needs sage.libs.pari
        sage: Q1.has_equivalent_Jordan_decomposition_at_prime(Q2, 2)
        False
        sage: Q1.has_equivalent_Jordan_decomposition_at_prime(Q2, 11)
        False
        sage: Q1.has_equivalent_Jordan_decomposition_at_prime(Q3, 2)
        False
        sage: Q1.has_equivalent_Jordan_decomposition_at_prime(Q3, 11)
        True
        sage: Q2.has_equivalent_Jordan_decomposition_at_prime(Q3, 2)
        True
        sage: Q2.has_equivalent_Jordan_decomposition_at_prime(Q3, 11)
        False
    """
    # Sanity Checks
    # if not isinstance(other, QuadraticForm):
    if not isinstance(other, type(self)):
        raise TypeError("the first argument must be of type QuadraticForm")
    if not is_prime(p):
        raise TypeError("the second argument must be a prime number")

    # Get the relevant local normal forms quickly
    self_jordan = self.jordan_blocks_by_scale_and_unimodular(p, safe_flag=False)
    other_jordan = other.jordan_blocks_by_scale_and_unimodular(p, safe_flag=False)

    # Check for the same number of Jordan components
    if len(self_jordan) != len(other_jordan):
        return False

    # Deal with odd primes:  Check that the Jordan component scales, dimensions, and discriminants are the same
    if p != 2:
        for i in range(len(self_jordan)):
            if (self_jordan[i][0] != other_jordan[i][0]) \
               or (self_jordan[i][1].dim() != other_jordan[i][1].dim()) \
               or (legendre_symbol(self_jordan[i][1].det() * other_jordan[i][1].det(), p) != 1):
                return False

        # All tests passed for an odd prime.
        return True

    # For p = 2:  Check that all Jordan Invariants are the same.
    elif p == 2:

        # Useful definition
        t = len(self_jordan)          # Define t = Number of Jordan components

        # Check that all Jordan Invariants are the same (scale, dim, and norm)
        for i in range(t):
            if (self_jordan[i][0] != other_jordan[i][0]) \
               or (self_jordan[i][1].dim() != other_jordan[i][1].dim()) \
               or (valuation(GCD(self_jordan[i][1].coefficients()), p) != valuation(GCD(other_jordan[i][1].coefficients()), p)):
                return False

        # Use O'Meara's isometry test 93:29 on p277.
        # ------------------------------------------

        # List of norms, scales, and dimensions for each i
        scale_list = [ZZ(2)**self_jordan[i][0] for i in range(t)]
        norm_list = [ZZ(2)**(self_jordan[i][0] + valuation(GCD(self_jordan[i][1].coefficients()), 2)) for i in range(t)]
        dim_list = [(self_jordan[i][1].dim()) for i in range(t)]

        # List of Hessian determinants and Hasse invariants for each Jordan (sub)chain
        # (Note: This is not the same as O'Meara's Gram determinants, but ratios are the same!)  -- NOT SO GOOD...
        # But it matters in condition (ii), so we multiply all by 2 (instead of dividing by 2 since only square-factors matter, and it's easier.)
        j = 0
        self_chain_det_list = [self_jordan[j][1].Gram_det() * (scale_list[j]**dim_list[j])]
        other_chain_det_list = [other_jordan[j][1].Gram_det() * (scale_list[j]**dim_list[j])]
        self_hasse_chain_list = [self_jordan[j][1].scale_by_factor(ZZ(2)**self_jordan[j][0]).hasse_invariant__OMeara(2)]
        other_hasse_chain_list = [other_jordan[j][1].scale_by_factor(ZZ(2)**other_jordan[j][0]).hasse_invariant__OMeara(2)]

        for j in range(1, t):
            self_chain_det_list.append(self_chain_det_list[j-1] * self_jordan[j][1].Gram_det() * (scale_list[j]**dim_list[j]))
            other_chain_det_list.append(other_chain_det_list[j-1] * other_jordan[j][1].Gram_det() * (scale_list[j]**dim_list[j]))
            self_hasse_chain_list.append(self_hasse_chain_list[j-1]
                                         * hilbert_symbol(self_chain_det_list[j-1], self_jordan[j][1].Gram_det(), 2)
                                         * self_jordan[j][1].hasse_invariant__OMeara(2))
            other_hasse_chain_list.append(other_hasse_chain_list[j-1]
                                          * hilbert_symbol(other_chain_det_list[j-1], other_jordan[j][1].Gram_det(), 2)
                                          * other_jordan[j][1].hasse_invariant__OMeara(2))

        # SANITY CHECK -- check that the scale powers are strictly increasing
        for i in range(1, len(scale_list)):
            if scale_list[i - 1] >= scale_list[i]:
                raise RuntimeError("there is something wrong with the Jordan Decomposition ; the given scales are not strictly increasing")

        # Test O'Meara's two conditions
        for i in range(t - 1):

            # Condition (i): Check that their (unit) ratio is a square (but it suffices to check at most mod 8).
            modulus = norm_list[i] * norm_list[i+1] / (scale_list[i] ** 2)
            modulus = min(modulus, 8)
            if (modulus > 1) and (((self_chain_det_list[i] / other_chain_det_list[i]) % modulus) != 1):
                return False

            # Check O'Meara's condition (ii) when appropriate
            if norm_list[i + 1] % (4 * norm_list[i]) == 0:
                if self_hasse_chain_list[i] * hilbert_symbol(norm_list[i] * other_chain_det_list[i], -self_chain_det_list[i], 2) \
                   != other_hasse_chain_list[i] * hilbert_symbol(norm_list[i], -other_chain_det_list[i], 2):      # Nipp conditions
                    return False

        # All tests passed for the prime 2.
        return True

    else:
        raise TypeError("this should not have happened")


def is_rationally_isometric(self, other, return_matrix=False) -> bool | Any:
    """
    Determine if two regular quadratic forms over a number field are isometric.

    INPUT:

    - ``other`` -- a quadratic form over a number field

    - ``return_matrix`` -- boolean (default: ``False``); return
      the transformation matrix instead of a boolean; this is currently
      only implemented for forms over ``QQ``

    OUTPUT:

    - if ``return_matrix`` is ``False``: a boolean

    - if ``return_matrix`` is ``True``: either ``False`` or the
      transformation matrix

    EXAMPLES::

        sage: V = DiagonalQuadraticForm(QQ, [1, 1, 2])
        sage: W = DiagonalQuadraticForm(QQ, [2, 2, 2])
        sage: V.is_rationally_isometric(W)                                              # needs sage.libs.pari
        True

    ::

        sage: # needs sage.rings.number_field
        sage: x = polygen(ZZ, 'x')
        sage: K.<a> = NumberField(x^2 - 3)
        sage: V = QuadraticForm(K, 4, [1, 0, 0, 0, 2*a, 0, 0, a, 0, 2]); V
        Quadratic form in 4 variables over Number Field in a
         with defining polynomial x^2 - 3 with coefficients:
        [ 1 0 0 0 ]
        [ * 2*a 0 0 ]
        [ * * a 0 ]
        [ * * * 2 ]
        sage: W = QuadraticForm(K, 4, [1, 2*a, 4, 6, 3, 10, 2, 1, 2, 5]); W
        Quadratic form in 4 variables over Number Field in a
         with defining polynomial x^2 - 3 with coefficients:
        [ 1 2*a 4 6 ]
        [ * 3 10 2 ]
        [ * * 1 2 ]
        [ * * * 5 ]
        sage: V.is_rationally_isometric(W)
        False

    ::

        sage: # needs sage.rings.number_field
        sage: K.<a> = NumberField(x^4 + 2*x + 6)
        sage: V = DiagonalQuadraticForm(K, [a, 2, 3, 2, 1]); V
        Quadratic form in 5 variables over Number Field in a
         with defining polynomial x^4 + 2*x + 6 with coefficients:
        [ a 0 0 0 0 ]
        [ * 2 0 0 0 ]
        [ * * 3 0 0 ]
        [ * * * 2 0 ]
        [ * * * * 1 ]
        sage: W = DiagonalQuadraticForm(K, [a, a, a, 2, 1]); W
        Quadratic form in 5 variables over Number Field in a
         with defining polynomial x^4 + 2*x + 6 with coefficients:
        [ a 0 0 0 0 ]
        [ * a 0 0 0 ]
        [ * * a 0 0 ]
        [ * * * 2 0 ]
        [ * * * * 1 ]
        sage: V.is_rationally_isometric(W)
        False

    ::

        sage: # needs sage.rings.number_field
        sage: K.<a> = NumberField(x^2 - 3)
        sage: V = DiagonalQuadraticForm(K, [-1, a, -2*a])
        sage: W = DiagonalQuadraticForm(K, [-1, -a, 2*a])
        sage: V.is_rationally_isometric(W)
        True

        sage: # needs sage.rings.number_field
        sage: V = DiagonalQuadraticForm(QQ, [1, 1, 2])
        sage: W = DiagonalQuadraticForm(QQ, [2, 2, 2])
        sage: T = V.is_rationally_isometric(W, True); T
        [   0    0    1]
        [-1/2 -1/2    0]
        [ 1/2 -1/2    0]
        sage: V.Gram_matrix() == T.transpose() * W.Gram_matrix() * T
        True

        sage: T = W.is_rationally_isometric(V, True); T                                 # needs sage.rings.number_field
        [ 0 -1  1]
        [ 0 -1 -1]
        [ 1  0  0]
        sage: W.Gram_matrix() == T.T * V.Gram_matrix() * T                              # needs sage.rings.number_field
        True

    ::

        sage: L = QuadraticForm(QQ, 3, [2, 2, 0, 2, 2, 5])
        sage: M = QuadraticForm(QQ, 3, [2, 2, 0, 3, 2, 3])
        sage: L.is_rationally_isometric(M, True)                                        # needs sage.libs.pari
        False

    ::

        sage: A = DiagonalQuadraticForm(QQ, [1, 5])
        sage: B = QuadraticForm(QQ, 2, [1, 12, 81])
        sage: T = A.is_rationally_isometric(B, True); T                                 # needs sage.libs.pari
        [  1  -2]
        [  0 1/3]
        sage: A.Gram_matrix() == T.T * B.Gram_matrix() * T                              # needs sage.libs.pari
        True

    ::

        sage: C = DiagonalQuadraticForm(QQ, [1, 5, 9])
        sage: D = DiagonalQuadraticForm(QQ, [6, 30, 1])
        sage: T = C.is_rationally_isometric(D, True); T                                 # needs sage.libs.pari
        [   0 -5/6  1/2]
        [   0  1/6  1/2]
        [  -1    0    0]
        sage: C.Gram_matrix() == T.T * D.Gram_matrix() * T                              # needs sage.libs.pari
        True

    ::

        sage: E = DiagonalQuadraticForm(QQ, [1, 1])
        sage: F = QuadraticForm(QQ, 2, [17, 94, 130])
        sage: T = F.is_rationally_isometric(E, True); T                                 # needs sage.libs.pari
        [     -4 -189/17]
        [     -1  -43/17]
        sage: F.Gram_matrix() == T.T * E.Gram_matrix() * T                              # needs sage.libs.pari
        True

    TESTS::

        sage: # needs sage.rings.number_field
        sage: K.<a> = QuadraticField(3)
        sage: V = DiagonalQuadraticForm(K, [1, 2])
        sage: W = DiagonalQuadraticForm(K, [1, 0])
        sage: V.is_rationally_isometric(W)
        Traceback (most recent call last):
        ...
        NotImplementedError: this only tests regular forms

    Forms must have the same base ring otherwise a :exc:`TypeError` is raised::

        sage: # needs sage.rings.number_field
        sage: K1.<a> = QuadraticField(5)
        sage: K2.<b> = QuadraticField(7)
        sage: V = DiagonalQuadraticForm(K1, [1, a])
        sage: W = DiagonalQuadraticForm(K2, [1, b])
        sage: V.is_rationally_isometric(W)
        Traceback (most recent call last):
        ...
        TypeError: forms must have the same base ring.

    Forms which have different dimension are not isometric::

        sage: W = DiagonalQuadraticForm(QQ, [1, 2])
        sage: V = DiagonalQuadraticForm(QQ, [1, 1, 1])
        sage: V.is_rationally_isometric(W)
        False

    Forms whose determinants do not differ by a square in the base field are not isometric::

        sage: # needs sage.rings.number_field
        sage: K.<a> = NumberField(x^2 - 3)
        sage: V = DiagonalQuadraticForm(K, [-1, a, -2*a])
        sage: W = DiagonalQuadraticForm(K, [-1, a, 2*a])
        sage: V.is_rationally_isometric(W)
        False

    ::

        sage: # needs sage.libs.gsl sage.rings.number_field
        sage: K.<a> = NumberField(x^5 - x + 2, 'a')
        sage: Q = QuadraticForm(K, 3, [a, 1, 0, -a**2, -a**3, -1])
        sage: m = Q.matrix()
        sage: for _ in range(5):
        ....:     t = random_matrix(ZZ, 3, algorithm='unimodular')
        ....:     m2 = t*m*t.transpose()
        ....:     Q2 = QuadraticForm(K, 3, [m2[i,j] / (2 if i==j else 1)
        ....:                               for i in range(3) for j in range(i,3)])
        ....:     print(Q.is_rationally_isometric(Q2))
        True
        True
        True
        True
        True
    """
    if self.Gram_det() == 0 or other.Gram_det() == 0:
        raise NotImplementedError("this only tests regular forms")

    if self.base_ring() != other.base_ring():
        raise TypeError("forms must have the same base ring.")

    if self.dim() != other.dim():
        return False

    if not (self.Gram_det() * other.Gram_det()).is_square():
        return False

    L1 = self.Gram_det().support()
    L2 = other.Gram_det().support()

    for p in set().union(L1, L2):
        if self.hasse_invariant(p) != other.hasse_invariant(p):
            return False

    if self.base_ring() == QQ:
        if self.signature() != other.signature():
            return False
    else:

        M = self.rational_diagonal_form().Gram_matrix_rational()
        N = other.rational_diagonal_form().Gram_matrix_rational()
        K = self.base_ring()

        Mentries = M.diagonal()
        Nentries = N.diagonal()

        for emb in K.real_embeddings():

            Mpos = 0
            for x in Mentries:
                Mpos += emb(x) >= 0

            Npos = 0
            for x in Nentries:
                Npos += emb(x) >= 0

            if Npos != Mpos:
                return False

    if not return_matrix:
        return True

    # Ensure that both quadratic forms are diagonal.
    Q, q_diagonal_transform = self.rational_diagonal_form(True)
    F, f_diagonal_transform = other.rational_diagonal_form(True)

    # Call the method that does all the work to compute the transformation.
    transform = _diagonal_isometry(Q, F)

    return f_diagonal_transform * transform * q_diagonal_transform.inverse()


def _diagonal_isometry(V, W):
    r"""
    Given two diagonal, rationally equivalent quadratic forms, computes a
    transition matrix mapping from one to the other.

    .. NOTE::

        This function is an auxiliary method of ``isometry``, which is
        the method that should be called as it performs error-checking
        that is not present in this function.

    INPUT:

    - ``V`` -- a diagonal quadratic form
    - ``W`` -- a diagonal quadratic form

    OUTPUT:

    - A matrix ``T`` representing the isometry transformation, such that if
      ``VM`` is the gram matrix of ``V`` and ``WM`` is the gram matrix of
      ``W``, then ``VM == T.transpose() * WM * T`` yields ``True``.

    EXAMPLES::

        sage: from sage.quadratic_forms.quadratic_form__equivalence_testing import _diagonal_isometry

        sage: Q = DiagonalQuadraticForm(QQ, [1, 2, 4])
        sage: F = DiagonalQuadraticForm(QQ, [2, 2, 2])

        sage: T = _diagonal_isometry(Q, F); T                                           # needs sage.libs.pari
        [   0    1    0]
        [-1/2    0    1]
        [ 1/2    0    1]
        sage: Q.Gram_matrix() == T.T * F.Gram_matrix() * T                              # needs sage.libs.pari
        True

        sage: T = _diagonal_isometry(F, Q); T                                           # needs sage.libs.pari
        [   0   -1   -1]
        [   1    0    0]
        [   0 -1/2  1/2]
        sage: F.Gram_matrix() == T.T * Q.Gram_matrix() * T                              # needs sage.libs.pari
        True
    """
    import copy
    from sage.quadratic_forms.quadratic_form import DiagonalQuadraticForm
    from sage.matrix.constructor import Matrix
    from sage.modules.free_module_element import vector

    # We need to modify V and W, so copy them into Q and F respectively.
    Q, F = copy.deepcopy(V), copy.deepcopy(W)
    # Let FM denote the Gram matrix of F.
    FM = F.Gram_matrix()
    n = Q.dim()

    # This matrix represents a new basis for W, where the columns of the
    # matrix are the vectors of the basis. We initialize it to the standard basis.
    change_of_basis_matrix = Matrix.identity(QQ, n)

    # The goal of this loop is to obtain a new basis for W such that the
    # Gram matrix of V with respect to the standard basis equals the Gram matrix
    # of W with respect to the new basis.
    for i in range(n):
        # If the first terms are not equal...
        if Q.Gram_matrix()[0][0] != F.Gram_matrix()[0][0]:
            # Find a vector w in F such that F(w) equals the first term of Q.
            w = F.solve(Q.Gram_matrix()[0][0])
            w = vector(QQ, i * [0] + w.list())

            # We want to extend the basis of W to include the vector w.
            # Find a non-fixed vector in the current basis to replace by w.
            j = i
            # The new set of vectors must still be linearly independent (i.e. the matrix is non-singular).
            while True:
                temp_matrix = Matrix(change_of_basis_matrix)
                temp_matrix.set_column(j, change_of_basis_matrix * w)
                if not temp_matrix.is_singular():
                    break
                j += 1

            change_of_basis_matrix = temp_matrix

            # We want to fix w to be the basis vector at position i, so swap it with whatever is already there.
            col = change_of_basis_matrix.column(i)
            change_of_basis_matrix.set_column(i, change_of_basis_matrix.column(j))
            change_of_basis_matrix.set_column(j, col)

            # Orthogonalize the basis.
            change_of_basis_matrix = _gram_schmidt(change_of_basis_matrix, i, W.bilinear_map)

            # Obtain the diagonal gram matrix of F.
            FM = W(change_of_basis_matrix).Gram_matrix_rational()

        # Now we have that QM[0][0] == FM[0][0] where QM and FM are the Gram matrices
        # of Q and F respectively. We remove the first variable from each form and continue.
        F = DiagonalQuadraticForm(F.base_ring(), FM.diagonal())
        F = F.extract_variables(range(i + 1, F.dim()))
        Q = Q.extract_variables(range(1, Q.dim()))

    return change_of_basis_matrix


def _gram_schmidt(m, fixed_vector_index, inner_product):
    r"""
    Orthogonalize a set of vectors, starting at a fixed vector, with respect to a given
    inner product.

    INPUT:

    - ``m`` -- a square matrix whose columns represent vectors
    - ``fixed_vector_index`` -- any vectors preceding the vector (i.e. to its left)
        at this index are not changed.
    - ``inner_product`` -- a function that takes two vector arguments and returns a scalar,
        representing an inner product

    OUTPUT: a matrix consisting of orthogonal columns with respect to the given inner product

    EXAMPLES::

        sage: from sage.quadratic_forms.quadratic_form__equivalence_testing import _gram_schmidt
        sage: Q = QuadraticForm(QQ, 3, [1, 2, 2, 2, 1, 3]); Q
        Quadratic form in 3 variables over Rational Field with coefficients:
        [ 1 2 2 ]
        [ * 2 1 ]
        [ * * 3 ]
        sage: QM = Q.Gram_matrix(); QM
        [  1   1   1]
        [  1   2 1/2]
        [  1 1/2   3]
        sage: std_basis = matrix.identity(3)
        sage: ortho_basis = _gram_schmidt(std_basis, 0, Q.bilinear_map); ortho_basis
        [   1   -1 -3/2]
        [   0    1  1/2]
        [   0    0    1]
        sage: Q(ortho_basis).Gram_matrix_rational()
        [  1   0   0]
        [  0   1   0]
        [  0   0 7/4]
        sage: v1 = ortho_basis.column(0); v2 = ortho_basis.column(1); v3 = ortho_basis.column(2);
        sage: Q.bilinear_map(v1, v2) == 0
        True
        sage: Q.bilinear_map(v1, v3) == 0
        True
        sage: Q.bilinear_map(v2, v3) == 0
        True
    """
    from sage.matrix.constructor import column_matrix

    n = m.dimensions()[0]
    vectors = [m.column(i) for i in range(n)]

    for i in range(fixed_vector_index, n):
        for j in range(i + 1, n):
            vectors[j] = vectors[j] - (inner_product(vectors[j], vectors[i]) / inner_product(vectors[i], vectors[i])) * vectors[i]

    return column_matrix(vectors)



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/acb.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void acb_init(acb_t x) noexcept
    void acb_clear(acb_t x) noexcept
    acb_ptr _acb_vec_init(slong n) noexcept
    void _acb_vec_clear(acb_ptr v, slong n) noexcept
    slong acb_allocated_bytes(const acb_t x) noexcept
    slong _acb_vec_allocated_bytes(acb_srcptr vec, slong len) noexcept
    double _acb_vec_estimate_allocated_bytes(slong len, slong prec) noexcept
    void acb_zero(acb_t z) noexcept
    void acb_one(acb_t z) noexcept
    void acb_onei(acb_t z) noexcept
    void acb_set(acb_t z, const acb_t x) noexcept
    void acb_set_ui(acb_t z, ulong x) noexcept
    void acb_set_si(acb_t z, slong x) noexcept
    void acb_set_d(acb_t z, double x) noexcept
    void acb_set_fmpz(acb_t z, const fmpz_t x) noexcept
    void acb_set_arb(acb_t z, const arb_t c) noexcept
    void acb_set_si_si(acb_t z, slong x, slong y) noexcept
    void acb_set_d_d(acb_t z, double x, double y) noexcept
    void acb_set_fmpz_fmpz(acb_t z, const fmpz_t x, const fmpz_t y) noexcept
    void acb_set_arb_arb(acb_t z, const arb_t x, const arb_t y) noexcept
    void acb_set_fmpq(acb_t z, const fmpq_t x, slong prec) noexcept
    void acb_set_round(acb_t z, const acb_t x, slong prec) noexcept
    void acb_set_round_fmpz(acb_t z, const fmpz_t x, slong prec) noexcept
    void acb_set_round_arb(acb_t z, const arb_t x, slong prec) noexcept
    void acb_swap(acb_t z, acb_t x) noexcept
    void acb_add_error_arf(acb_t x, const arf_t err) noexcept
    void acb_add_error_mag(acb_t x, const mag_t err) noexcept
    void acb_add_error_arb(acb_t x, const arb_t err) noexcept
    void acb_get_mid(acb_t m, const acb_t x) noexcept
    void acb_print(const acb_t x) noexcept
    void acb_fprint(FILE * file, const acb_t x) noexcept
    void acb_printd(const acb_t x, slong digits) noexcept
    void acb_fprintd(FILE * file, const acb_t x, slong digits) noexcept
    void acb_printn(const acb_t x, slong digits, ulong flags) noexcept
    void acb_fprintn(FILE * file, const acb_t x, slong digits, ulong flags) noexcept
    void acb_randtest(acb_t z, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void acb_randtest_special(acb_t z, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void acb_randtest_precise(acb_t z, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void acb_randtest_param(acb_t z, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void acb_urandom(acb_t z, flint_rand_t state, slong prec) noexcept
    bint acb_is_zero(const acb_t z) noexcept
    bint acb_is_one(const acb_t z) noexcept
    bint acb_is_finite(const acb_t z) noexcept
    bint acb_is_exact(const acb_t z) noexcept
    bint acb_is_int(const acb_t z) noexcept
    bint acb_is_int_2exp_si(const acb_t x, slong e) noexcept
    bint acb_equal(const acb_t x, const acb_t y) noexcept
    bint acb_equal_si(const acb_t x, slong y) noexcept
    bint acb_eq(const acb_t x, const acb_t y) noexcept
    bint acb_ne(const acb_t x, const acb_t y) noexcept
    bint acb_overlaps(const acb_t x, const acb_t y) noexcept
    void acb_union(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_get_abs_ubound_arf(arf_t u, const acb_t z, slong prec) noexcept
    void acb_get_abs_lbound_arf(arf_t u, const acb_t z, slong prec) noexcept
    void acb_get_rad_ubound_arf(arf_t u, const acb_t z, slong prec) noexcept
    void acb_get_mag(mag_t u, const acb_t x) noexcept
    void acb_get_mag_lower(mag_t u, const acb_t x) noexcept
    bint acb_contains_fmpq(const acb_t x, const fmpq_t y) noexcept
    bint acb_contains_fmpz(const acb_t x, const fmpz_t y) noexcept
    bint acb_contains(const acb_t x, const acb_t y) noexcept
    bint acb_contains_zero(const acb_t x) noexcept
    bint acb_contains_int(const acb_t x) noexcept
    bint acb_contains_interior(const acb_t x, const acb_t y) noexcept
    slong acb_rel_error_bits(const acb_t x) noexcept
    slong acb_rel_accuracy_bits(const acb_t x) noexcept
    slong acb_rel_one_accuracy_bits(const acb_t x) noexcept
    slong acb_bits(const acb_t x) noexcept
    void acb_indeterminate(acb_t x) noexcept
    void acb_trim(acb_t y, const acb_t x) noexcept
    bint acb_is_real(const acb_t x) noexcept
    int acb_get_unique_fmpz(fmpz_t z, const acb_t x) noexcept
    void acb_get_real(arb_t re, const acb_t z) noexcept
    void acb_get_imag(arb_t im, const acb_t z) noexcept
    void acb_arg(arb_t r, const acb_t z, slong prec) noexcept
    void acb_abs(arb_t r, const acb_t z, slong prec) noexcept
    void acb_sgn(acb_t r, const acb_t z, slong prec) noexcept
    void acb_csgn(arb_t r, const acb_t z) noexcept
    void acb_neg(acb_t z, const acb_t x) noexcept
    void acb_neg_round(acb_t z, const acb_t x, slong prec) noexcept
    void acb_conj(acb_t z, const acb_t x) noexcept
    void acb_add_ui(acb_t z, const acb_t x, ulong y, slong prec) noexcept
    void acb_add_si(acb_t z, const acb_t x, slong y, slong prec) noexcept
    void acb_add_fmpz(acb_t z, const acb_t x, const fmpz_t y, slong prec) noexcept
    void acb_add_arb(acb_t z, const acb_t x, const arb_t y, slong prec) noexcept
    void acb_add(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_sub_ui(acb_t z, const acb_t x, ulong y, slong prec) noexcept
    void acb_sub_si(acb_t z, const acb_t x, slong y, slong prec) noexcept
    void acb_sub_fmpz(acb_t z, const acb_t x, const fmpz_t y, slong prec) noexcept
    void acb_sub_arb(acb_t z, const acb_t x, const arb_t y, slong prec) noexcept
    void acb_sub(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_mul_onei(acb_t z, const acb_t x) noexcept
    void acb_div_onei(acb_t z, const acb_t x) noexcept
    void acb_mul_i_pow_si(acb_t z, const acb_t x, slong k) noexcept
    void acb_mul_ui(acb_t z, const acb_t x, ulong y, slong prec) noexcept
    void acb_mul_si(acb_t z, const acb_t x, slong y, slong prec) noexcept
    void acb_mul_fmpz(acb_t z, const acb_t x, const fmpz_t y, slong prec) noexcept
    void acb_mul_arb(acb_t z, const acb_t x, const arb_t y, slong prec) noexcept
    void acb_mul(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_mul_2exp_si(acb_t z, const acb_t x, slong e) noexcept
    void acb_mul_2exp_fmpz(acb_t z, const acb_t x, const fmpz_t e) noexcept
    void acb_sqr(acb_t z, const acb_t x, slong prec) noexcept
    void acb_cube(acb_t z, const acb_t x, slong prec) noexcept
    void acb_addmul(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_addmul_ui(acb_t z, const acb_t x, ulong y, slong prec) noexcept
    void acb_addmul_si(acb_t z, const acb_t x, slong y, slong prec) noexcept
    void acb_addmul_fmpz(acb_t z, const acb_t x, const fmpz_t y, slong prec) noexcept
    void acb_addmul_arb(acb_t z, const acb_t x, const arb_t y, slong prec) noexcept
    void acb_submul(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_submul_ui(acb_t z, const acb_t x, ulong y, slong prec) noexcept
    void acb_submul_si(acb_t z, const acb_t x, slong y, slong prec) noexcept
    void acb_submul_fmpz(acb_t z, const acb_t x, const fmpz_t y, slong prec) noexcept
    void acb_submul_arb(acb_t z, const acb_t x, const arb_t y, slong prec) noexcept
    void acb_inv(acb_t z, const acb_t x, slong prec) noexcept
    void acb_div_ui(acb_t z, const acb_t x, ulong y, slong prec) noexcept
    void acb_div_si(acb_t z, const acb_t x, slong y, slong prec) noexcept
    void acb_div_fmpz(acb_t z, const acb_t x, const fmpz_t y, slong prec) noexcept
    void acb_div_arb(acb_t z, const acb_t x, const arb_t y, slong prec) noexcept
    void acb_div(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_dot_precise(acb_t res, const acb_t s, int subtract, acb_srcptr x, slong xstep, acb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void acb_dot_simple(acb_t res, const acb_t s, int subtract, acb_srcptr x, slong xstep, acb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void acb_dot(acb_t res, const acb_t s, int subtract, acb_srcptr x, slong xstep, acb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void acb_approx_dot(acb_t res, const acb_t s, int subtract, acb_srcptr x, slong xstep, acb_srcptr y, slong ystep, slong len, slong prec) noexcept
    void acb_dot_ui(acb_t res, const acb_t initial, int subtract, acb_srcptr x, slong xstep, const ulong * y, slong ystep, slong len, slong prec) noexcept
    void acb_dot_si(acb_t res, const acb_t initial, int subtract, acb_srcptr x, slong xstep, const slong * y, slong ystep, slong len, slong prec) noexcept
    void acb_dot_uiui(acb_t res, const acb_t initial, int subtract, acb_srcptr x, slong xstep, const ulong * y, slong ystep, slong len, slong prec) noexcept
    void acb_dot_siui(acb_t res, const acb_t initial, int subtract, acb_srcptr x, slong xstep, const ulong * y, slong ystep, slong len, slong prec) noexcept
    void acb_dot_fmpz(acb_t res, const acb_t initial, int subtract, acb_srcptr x, slong xstep, const fmpz * y, slong ystep, slong len, slong prec) noexcept
    void acb_const_pi(acb_t y, slong prec) noexcept
    void acb_sqrt(acb_t r, const acb_t z, slong prec) noexcept
    void acb_sqrt_analytic(acb_t r, const acb_t z, int analytic, slong prec) noexcept
    void acb_rsqrt(acb_t r, const acb_t z, slong prec) noexcept
    void acb_rsqrt_analytic(acb_t r, const acb_t z, int analytic, slong prec) noexcept
    void acb_sqrts(acb_t y1, acb_t y2, const acb_t x, slong prec) noexcept
    void acb_quadratic_roots_fmpz(acb_t r1, acb_t r2, const fmpz_t a, const fmpz_t b, const fmpz_t c, slong prec) noexcept
    void acb_root_ui(acb_t r, const acb_t z, ulong k, slong prec) noexcept
    void acb_pow_fmpz(acb_t y, const acb_t b, const fmpz_t e, slong prec) noexcept
    void acb_pow_ui(acb_t y, const acb_t b, ulong e, slong prec) noexcept
    void acb_pow_si(acb_t y, const acb_t b, slong e, slong prec) noexcept
    void acb_pow_arb(acb_t z, const acb_t x, const arb_t y, slong prec) noexcept
    void acb_pow(acb_t z, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_pow_analytic(acb_t r, const acb_t x, const acb_t y, int analytic, slong prec) noexcept
    void acb_unit_root(acb_t res, ulong order, slong prec) noexcept
    void acb_exp(acb_t y, const acb_t z, slong prec) noexcept
    void acb_exp_pi_i(acb_t y, const acb_t z, slong prec) noexcept
    void acb_exp_invexp(acb_t s, acb_t t, const acb_t z, slong prec) noexcept
    void acb_expm1(acb_t res, const acb_t z, slong prec) noexcept
    void acb_log(acb_t y, const acb_t z, slong prec) noexcept
    void acb_log_analytic(acb_t r, const acb_t z, int analytic, slong prec) noexcept
    void acb_log1p(acb_t z, const acb_t x, slong prec) noexcept
    void acb_sin(acb_t s, const acb_t z, slong prec) noexcept
    void acb_cos(acb_t c, const acb_t z, slong prec) noexcept
    void acb_sin_cos(acb_t s, acb_t c, const acb_t z, slong prec) noexcept
    void acb_tan(acb_t s, const acb_t z, slong prec) noexcept
    void acb_cot(acb_t s, const acb_t z, slong prec) noexcept
    void acb_sin_pi(acb_t s, const acb_t z, slong prec) noexcept
    void acb_cos_pi(acb_t s, const acb_t z, slong prec) noexcept
    void acb_sin_cos_pi(acb_t s, acb_t c, const acb_t z, slong prec) noexcept
    void acb_tan_pi(acb_t s, const acb_t z, slong prec) noexcept
    void acb_cot_pi(acb_t s, const acb_t z, slong prec) noexcept
    void acb_sec(acb_t res, const acb_t z, slong prec) noexcept
    void acb_csc(acb_t res, const acb_t z, slong prec) noexcept
    void acb_csc_pi(acb_t res, const acb_t z, slong prec) noexcept
    void acb_sinc(acb_t s, const acb_t z, slong prec) noexcept
    void acb_sinc_pi(acb_t s, const acb_t z, slong prec) noexcept
    void acb_asin(acb_t res, const acb_t z, slong prec) noexcept
    void acb_acos(acb_t res, const acb_t z, slong prec) noexcept
    void acb_atan(acb_t res, const acb_t z, slong prec) noexcept
    void acb_sinh(acb_t s, const acb_t z, slong prec) noexcept
    void acb_cosh(acb_t c, const acb_t z, slong prec) noexcept
    void acb_sinh_cosh(acb_t s, acb_t c, const acb_t z, slong prec) noexcept
    void acb_tanh(acb_t s, const acb_t z, slong prec) noexcept
    void acb_coth(acb_t s, const acb_t z, slong prec) noexcept
    void acb_sech(acb_t res, const acb_t z, slong prec) noexcept
    void acb_csch(acb_t res, const acb_t z, slong prec) noexcept
    void acb_asinh(acb_t res, const acb_t z, slong prec) noexcept
    void acb_acosh(acb_t res, const acb_t z, slong prec) noexcept
    void acb_atanh(acb_t res, const acb_t z, slong prec) noexcept
    void acb_lambertw_asymp(acb_t res, const acb_t z, const fmpz_t k, slong L, slong M, slong prec) noexcept
    int acb_lambertw_check_branch(const acb_t w, const fmpz_t k, slong prec) noexcept
    void acb_lambertw_bound_deriv(mag_t res, const acb_t z, const acb_t ez1, const fmpz_t k) noexcept
    void acb_lambertw(acb_t res, const acb_t z, const fmpz_t k, int flags, slong prec) noexcept
    void acb_rising_ui(acb_t z, const acb_t x, ulong n, slong prec) noexcept
    void acb_rising(acb_t z, const acb_t x, const acb_t n, slong prec) noexcept
    void acb_gamma(acb_t y, const acb_t x, slong prec) noexcept
    void acb_rgamma(acb_t y, const acb_t x, slong prec) noexcept
    void acb_lgamma(acb_t y, const acb_t x, slong prec) noexcept
    void acb_digamma(acb_t y, const acb_t x, slong prec) noexcept
    void acb_log_sin_pi(acb_t res, const acb_t z, slong prec) noexcept
    void acb_polygamma(acb_t res, const acb_t s, const acb_t z, slong prec) noexcept
    void acb_barnes_g(acb_t res, const acb_t z, slong prec) noexcept
    void acb_log_barnes_g(acb_t res, const acb_t z, slong prec) noexcept
    void acb_zeta(acb_t z, const acb_t s, slong prec) noexcept
    void acb_hurwitz_zeta(acb_t z, const acb_t s, const acb_t a, slong prec) noexcept
    void acb_bernoulli_poly_ui(acb_t res, ulong n, const acb_t x, slong prec) noexcept
    void acb_polylog(acb_t w, const acb_t s, const acb_t z, slong prec) noexcept
    void acb_polylog_si(acb_t w, slong s, const acb_t z, slong prec) noexcept
    void acb_agm1(acb_t m, const acb_t z, slong prec) noexcept
    void acb_agm1_cpx(acb_ptr m, const acb_t z, slong len, slong prec) noexcept
    void acb_agm(acb_t m, const acb_t x, const acb_t y, slong prec) noexcept
    void acb_chebyshev_t_ui(acb_t a, ulong n, const acb_t x, slong prec) noexcept
    void acb_chebyshev_u_ui(acb_t a, ulong n, const acb_t x, slong prec) noexcept
    void acb_chebyshev_t2_ui(acb_t a, acb_t b, ulong n, const acb_t x, slong prec) noexcept
    void acb_chebyshev_u2_ui(acb_t a, acb_t b, ulong n, const acb_t x, slong prec) noexcept
    void acb_real_abs(acb_t res, const acb_t z, int analytic, slong prec) noexcept
    void acb_real_sgn(acb_t res, const acb_t z, int analytic, slong prec) noexcept
    void acb_real_heaviside(acb_t res, const acb_t z, int analytic, slong prec) noexcept
    void acb_real_floor(acb_t res, const acb_t z, int analytic, slong prec) noexcept
    void acb_real_ceil(acb_t res, const acb_t z, int analytic, slong prec) noexcept
    void acb_real_max(acb_t res, const acb_t x, const acb_t y, int analytic, slong prec) noexcept
    void acb_real_min(acb_t res, const acb_t x, const acb_t y, int analytic, slong prec) noexcept
    void acb_real_sqrtpos(acb_t res, const acb_t z, int analytic, slong prec) noexcept
    void _acb_vec_zero(acb_ptr A, slong n) noexcept
    bint _acb_vec_is_zero(acb_srcptr vec, slong len) noexcept
    bint _acb_vec_is_real(acb_srcptr v, slong len) noexcept
    bint _acb_vec_is_finite(acb_srcptr vec, slong len) noexcept
    bint _acb_vec_equal(acb_srcptr vec1, acb_srcptr vec2, slong len) noexcept
    bint _acb_vec_overlaps(acb_srcptr vec1, acb_srcptr vec2, slong len) noexcept
    bint _acb_vec_contains(acb_srcptr vec1, acb_srcptr vec2, slong len) noexcept
    void _acb_vec_set(acb_ptr res, acb_srcptr vec, slong len) noexcept
    void _acb_vec_set_round(acb_ptr res, acb_srcptr vec, slong len, slong prec) noexcept
    void _acb_vec_swap(acb_ptr vec1, acb_ptr vec2, slong len) noexcept
    void _acb_vec_get_real(arb_ptr re, acb_srcptr vec, slong len) noexcept
    void _acb_vec_get_imag(arb_ptr im, acb_srcptr vec, slong len) noexcept
    void _acb_vec_set_real_imag(acb_ptr vec, arb_srcptr re, arb_srcptr im, slong len) noexcept
    void _acb_vec_neg(acb_ptr res, acb_srcptr vec, slong len) noexcept
    void _acb_vec_add(acb_ptr res, acb_srcptr vec1, acb_srcptr vec2, slong len, slong prec) noexcept
    void _acb_vec_sub(acb_ptr res, acb_srcptr vec1, acb_srcptr vec2, slong len, slong prec) noexcept
    void _acb_vec_scalar_submul(acb_ptr res, acb_srcptr vec, slong len, const acb_t c, slong prec) noexcept
    void _acb_vec_scalar_addmul(acb_ptr res, acb_srcptr vec, slong len, const acb_t c, slong prec) noexcept
    void _acb_vec_scalar_mul(acb_ptr res, acb_srcptr vec, slong len, const acb_t c, slong prec) noexcept
    void _acb_vec_scalar_mul_ui(acb_ptr res, acb_srcptr vec, slong len, ulong c, slong prec) noexcept
    void _acb_vec_scalar_mul_2exp_si(acb_ptr res, acb_srcptr vec, slong len, slong c) noexcept
    void _acb_vec_scalar_mul_onei(acb_ptr res, acb_srcptr vec, slong len) noexcept
    void _acb_vec_scalar_div_ui(acb_ptr res, acb_srcptr vec, slong len, ulong c, slong prec) noexcept
    void _acb_vec_scalar_div(acb_ptr res, acb_srcptr vec, slong len, const acb_t c, slong prec) noexcept
    void _acb_vec_scalar_mul_arb(acb_ptr res, acb_srcptr vec, slong len, const arb_t c, slong prec) noexcept
    void _acb_vec_scalar_div_arb(acb_ptr res, acb_srcptr vec, slong len, const arb_t c, slong prec) noexcept
    void _acb_vec_scalar_mul_fmpz(acb_ptr res, acb_srcptr vec, slong len, const fmpz_t c, slong prec) noexcept
    void _acb_vec_scalar_div_fmpz(acb_ptr res, acb_srcptr vec, slong len, const fmpz_t c, slong prec) noexcept
    void _acb_vec_sqr(acb_ptr res, acb_srcptr vec, slong len, slong prec) noexcept
    slong _acb_vec_bits(acb_srcptr vec, slong len) noexcept
    void _acb_vec_set_powers(acb_ptr xs, const acb_t x, slong len, slong prec) noexcept
    void _acb_vec_unit_roots(acb_ptr z, slong order, slong len, slong prec) noexcept
    void _acb_vec_add_error_arf_vec(acb_ptr res, arf_srcptr err, slong len) noexcept
    void _acb_vec_add_error_mag_vec(acb_ptr res, mag_srcptr err, slong len) noexcept
    void _acb_vec_indeterminate(acb_ptr vec, slong len) noexcept
    void _acb_vec_trim(acb_ptr res, acb_srcptr vec, slong len) noexcept
    int _acb_vec_get_unique_fmpz_vec(fmpz * res,  acb_srcptr vec, slong len) noexcept
    void _acb_vec_sort_pretty(acb_ptr vec, slong len) noexcept
    void _acb_vec_printd(acb_srcptr vec, slong len, slong digits) noexcept
    void _acb_vec_printn(acb_srcptr vec, slong len, slong digits, ulong flags) noexcept

from .acb_macros cimport *



# sage_setup: distribution = sagemath-flint
# distutils: libraries = gmp
# distutils: extra_compile_args = -D_XPG6
# sage.doctest: needs sage.schemes
"""
`p`-adic distributions spaces

This module implements `p`-adic distributions, a `p`-adic Banach
space dual to locally analytic functions on a disc.

EXAMPLES::

    sage: D = OverconvergentDistributions(5, 7, 15)
    sage: v = D([7,14,21,28,35]); v
    (7 + O(7^5), 2*7 + O(7^4), 3*7 + O(7^3), 4*7 + O(7^2), O(7))

REFERENCES:

- [PS2011]_
"""

# ****************************************************************************
#       Copyright (C) 2012 Robert Pollack <rpollack@math.bu.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import operator

from sage.arith.misc import bernoulli
from sage.categories.fields import Fields
from sage.matrix.constructor import matrix
from sage.matrix.matrix cimport Matrix
from sage.misc.verbose import verbose
from sage.modular.pollack_stevens.sigma0 import Sigma0
from sage.rings.finite_rings.integer_mod_ring import Zmod
from sage.rings.infinity import Infinity
from sage.rings.integer cimport Integer
from sage.rings.integer_ring import ZZ
from sage.rings.padics.padic_generic import pAdicGeneric
from sage.rings.power_series_ring import PowerSeriesRing
from sage.rings.rational_field import QQ
from sage.structure.element cimport Element
from sage.structure.richcmp cimport richcmp_not_equal, rich_to_bool

# from sage.libs.flint.ulong_extras cimport *

cdef long overflow = 1 << (4 * sizeof(long) - 1)
cdef long underflow = -overflow
cdef long maxordp = (1L << (sizeof(long) * 8 - 2)) - 1


def get_dist_classes(p, prec_cap, base, symk, implementation):
    r"""
    Determine the element and action classes to be used for given inputs.

    INPUT:

    - ``p`` -- prime

    - ``prec_cap`` -- the `p`-adic precision cap

    - ``base`` -- the base ring

    - ``symk`` -- an element of Symk

    - ``implementation`` -- string; if not ``None``, override the
      automatic choice of implementation. May be 'long' or 'vector',
      otherwise raise a :exc:`NotImplementedError`.

    OUTPUT:

    - Either a Dist_vector and WeightKAction_vector, or a Dist_vector_long
      and WeightKAction_vector_long

    EXAMPLES::

        sage: D = OverconvergentDistributions(2, 3, 5); D # indirect doctest
        Space of 3-adic distributions with k=2 action and precision cap 5
    """
    if implementation is not None:
        if implementation == 'long':
            raise NotImplementedError('The optimized implementation -using longs- has been disabled and may return wrong results.')
        elif implementation == 'vector':
            return Dist_vector, WeightKAction_vector
        else:
            raise NotImplementedError('The implementation "%s" does not exist yet' % (implementation))

    return Dist_vector, WeightKAction_vector


cdef class Dist(ModuleElement):
    r"""
        The main `p`-adic distribution class, implemented as per the paper [PS2011]__.
    """
    def moment(self, n):
        r"""
        Return the `n`-th moment.

        INPUT:

        - ``n`` -- integer or slice, to be passed on to moments

        OUTPUT:

        - the `n`-th moment, or a list of moments in the case that `n`
          is a slice.

        EXAMPLES::

            sage: D = OverconvergentDistributions(4, 7, 10)
            sage: v = D([7,14,21,28,35])
            sage: v.moment(3)
            4*7 + O(7^2)
            sage: v.moment(0)
            7 + O(7^5)
        """
        return self.parent().prime() ** (self.ordp) * self._unscaled_moment(n)

    def moments(self):
        r"""
        Return the vector of moments.

        OUTPUT: the vector of moments

        EXAMPLES::

            sage: D = OverconvergentDistributions(4, 5, 10, base = Qp(5))
            sage: v = D([1,7,4,2,-1])
            sage: v = 1/5^3 * v
            sage: v
            5^-3 * (1 + O(5^5), 2 + 5 + O(5^4), 4 + O(5^3), 2 + O(5^2), 4 + O(5))
            sage: v.moments()
            (5^-3 + O(5^2), 2*5^-3 + 5^-2 + O(5), 4*5^-3 + O(5^0), 2*5^-3 + O(5^-1), 4*5^-3 + O(5^-2))
        """
        return self.parent().prime() ** (self.ordp) * self._moments

    cpdef normalize(self, include_zeroth_moment=True):
        r"""
        Normalize so that the precision of the `i`-th moment is `n-i`,
        where `n` is the number of moments.

        OUTPUT: normalized entries of the distribution

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15); D
            Space of 7-adic distributions with k=5 action and precision cap 15
            sage: v = D([1,2,3,4,5]); v
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: v.normalize()
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
        """
        raise NotImplementedError

    cdef long _relprec(self) noexcept:
        raise NotImplementedError

    cdef _unscaled_moment(self, long i):
        raise NotImplementedError

    cpdef long _ord_p(self) noexcept:
        r"""
        Return power of `p` by which the moments are shifted.

        .. NOTE::

            This is not necessarily the same as the valuation,
            since the moments could all be divisible by `p`.

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([7,14,21,28,35]); v
            (7 + O(7^5), 2*7 + O(7^4), 3*7 + O(7^3), 4*7 + O(7^2), O(7))
            sage: v._ord_p()
            0
        """
        return self.ordp

    def scale(self, left):
        r"""
        Scale the moments of the distribution by ``left``.

        INPUT:

        - ``left`` -- scalar

        OUTPUT: scales the moments by ``left``

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5]); v
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: v.scale(2)
            (2 + O(7^5), 4 + O(7^4), 6 + O(7^3), 1 + 7 + O(7^2), 3 + O(7))
        """
        # if isinstance(self, Dist_long) and isinstance(left, (Integer, pAdicCappedRelativeElement, pAdicCappedAbsoluteElement, pAdicFixedModElement)):
        #     return self._lmul_(left)
        R = left.parent()
        base = self.parent().base_ring()
        if base is R:
            return self._lmul_(left)
        elif base.has_coerce_map_from(R):
            return self._lmul_(base(left))
        else:
            from sage.categories.pushout import pushout
            new_base = pushout(base, R)
            V = self.parent().change_ring(new_base)
            scalar = new_base(left)
            return V([scalar * new_base(self.moment(i)) for i in range(self.precision_absolute())])

    def is_zero(self, p=None, M=None) -> bool:
        r"""
        Return ``True`` if the `i`-th moment is zero for all `i` (case ``M`` is
        ``None``) or zero modulo `p^{M-i}` for all `i` (when ``M`` is not
        ``None``).

        Note that some moments are not known to precision ``M``, in which
        case they are only checked to be equal to zero modulo the
        precision to which they are defined.

        INPUT:

        - ``p`` -- prime

        - ``M`` -- precision

        OUTPUT: boolean

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5]); v
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: v.is_zero()
            False
            sage: v = D(5*[0])
            sage: v.is_zero()
            True

        ::

            sage: D = Symk(0)
            sage: v = D([0])
            sage: v.is_zero(5,3)
            True
        """
        n = self.precision_relative()
        aprec = self.precision_absolute()
        if M is None:
            M = n
        # elif M > aprec: # DEBUG
        #     return False
        elif M < aprec:
            n -= (aprec - M)
            M -= self.ordp
        if p is None:
            p = self.parent().prime()
        cdef bint usearg = True
        if n == 0:
            return True
        else:
            try:
                z = self._unscaled_moment(0).is_zero(M)
            except TypeError:
                z = self._unscaled_moment(0).is_zero()
                use_arg = False
            if not z:
                return False
            for a in range(1, n):
                if usearg:
                    try:
                        z = self._unscaled_moment(a).is_zero(M - a)
                    except TypeError:
                        z = self._unscaled_moment(a).is_zero()
                        use_arg = False
                else:
                    z = self._unscaled_moment(a).is_zero()
                if not z:
                    return False
            return True

    def find_scalar(self, _other, p, M=None, check=True):
        r"""
        Return an ``alpha`` with ``other = self * alpha``, or raises
        a :exc:`ValueError`.

        It will also raise a :exc:`ValueError` if this distribution is zero.

        INPUT:

        - ``other`` -- another distribution

        - ``p`` -- an integral prime (only used if the parent is not a Symk)

        - ``M`` -- (default: ``None``) an integer, the relative precision
          to which the scalar must be determined

        - ``check`` -- boolean (default: ``True``); whether to validate
          that ``other`` is actually a multiple of this element

        OUTPUT: a scalar ``alpha`` with ``other = self * alpha``

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5])
            sage: w = D([3,6,9,12,15])
            sage: v.find_scalar(w,p=7)
            3 + O(7^5)
            sage: v.find_scalar(w,p=7,M=4)
            3 + O(7^4)

            sage: u = D([1,4,9,16,25])
            sage: v.find_scalar(u,p=7)
            Traceback (most recent call last):
            ...
            ValueError: not a scalar multiple
        """
        cdef Dist other = _other
        i = 0
        n = self.precision_relative()
        other_pr = other.precision_relative()
        if n == 0:
            raise ValueError("self is zero")
        verbose("n = %s" % n, level=2)
        verbose("moment 0", level=2)
        a = self._unscaled_moment(i)
        verbose("a = %s" % a, level = 2)
        padic = isinstance(a.parent(), pAdicGeneric)
        if self.parent().is_symk():
            while a == 0:
                if other._unscaled_moment(i) != 0:
                    raise ValueError("not a scalar multiple")
                i += 1
                verbose("moment %s" % i, level = 2)
                try:
                    a = self._unscaled_moment(i)
                except IndexError:
                    raise ValueError("self is zero")
            alpha = other._unscaled_moment(i) / a
            if check:
                i += 1
                while i < n:
                    verbose("comparing moment %s" % i, level = 2)
                    if alpha * self._unscaled_moment(i) != other._unscaled_moment(i):
                        raise ValueError("not a scalar multiple")
                    i += 1
        else:
            p = self.parent().prime()
            v = a.valuation(p)
            while v >= n - i:
                i += 1
                verbose("p moment %s" % i, level = 2)
                try:
                    a = self._unscaled_moment(i)
                except IndexError:
                    raise ValueError("self is zero")
                v = a.valuation(p)
            relprec = n - i - v
            #            verbose("p=%s, n-i=%s\nself.moment=%s, other.moment=%s" % (p, n-i, a, other._unscaled_moment(i)),level=2)
            # RP: This code was crashing because other may have too few moments -- so I added this bound with other's relative precision
            if padic:
                if i < other_pr:
                    alpha = (other._unscaled_moment(i) / a).add_bigoh(n - i)
                else:
                    alpha = (0 * a).add_bigoh(other_pr - i)
            else:
                if i < other_pr:
                    alpha = (other._unscaled_moment(i) / a) % p ** (n - i)
                else:
                    alpha = 0
            verbose("alpha = %s" % alpha, level = 2)
            # RP: This code was crashing because other may have too few moments -- so I added this bound with other's relative precision
            while i < other_pr - 1:
                i += 1
                verbose("comparing p moment %s" % i, level = 2)
                a = self._unscaled_moment(i)
                if check:
                    # verbose("self.moment=%s, other.moment=%s" % (a, other._unscaled_moment(i)))
                    if (padic and other._unscaled_moment(i) != alpha * a) or \
                       (not padic and other._unscaled_moment(i) % p ** (n - i) != alpha * a % p ** (n - i)):
                        raise ValueError("not a scalar multiple")
                v = a.valuation(p)
                if n - i - v > relprec:
                    verbose("Resetting alpha: relprec=%s, n-i=%s, v=%s" % (relprec, n - i, v), level = 2)
                    relprec = n - i - v
                    if padic:
                        alpha = (other._unscaled_moment(i) / a).add_bigoh(n - i)
                    else:
                        alpha = (other._unscaled_moment(i) / a) % p ** (n - i)
                    verbose("alpha=%s" % alpha, level = 2)
            if relprec < M:
                raise ValueError("result not determined to high enough precision")
        alpha = alpha * self.parent().prime() ** (other.ordp - self.ordp)
        verbose("alpha=%s" % alpha, level = 2)
        try:
            alpha = self.parent().base_ring()(alpha)
            if M is not None:
                alpha = alpha.add_bigoh(M)
        except (ValueError, AttributeError):
            pass
        return alpha

    def find_scalar_from_zeroth_moment(self, _other, p, M=None, check=True):
        r"""
        Return an ``alpha`` with ``other = self * alpha`` using only
        the zeroth moment, or raises a :exc:`ValueError`.

        It will also raise a :exc:`ValueError` if the zeroth moment of the
        distribution is zero.

        INPUT:

        - ``other`` -- another distribution

        - ``p`` -- an integral prime (only used if the parent is not a Symk)

        - ``M`` -- (default: ``None``) an integer, the relative precision
          to which the scalar must be determined

        - ``check`` -- boolean (default: ``True``); whether to validate
          that ``other`` is actually a multiple of this element

        OUTPUT: a scalar ``alpha`` with ``other = self * alpha``

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5])
            sage: w = D([3,6,9,12,15])
            sage: v.find_scalar_from_zeroth_moment(w,p=7)
            3 + O(7^5)
            sage: v.find_scalar_from_zeroth_moment(w,p=7,M=4)
            3 + O(7^4)

            sage: u = D([1,4,9,16,25])
            sage: v.find_scalar_from_zeroth_moment(u,p=7)
            Traceback (most recent call last):
            ...
            ValueError: not a scalar multiple
        """
        cdef Dist other = _other
        n = self.precision_relative()
        if n == 0:
            raise ValueError("zeroth moment is zero")
        verbose("n = %s" % n, level = 2)
        a = self.moment(0)
        if a.is_zero():
            raise ValueError("zeroth moment is zero")
        alpha = other.moment(0) / a
        if check:
            for i in range(1, n):
                verbose("comparing moment %s" % i, level = 2)
                if alpha * self.moment(i) != other.moment(i):
                    raise ValueError("not a scalar multiple")
        alpha = self.parent().base_ring()(alpha)
        if M is not None:
            try:
                absprec = alpha.precision_absolute()
                if absprec < M:
                    raise ValueError("result not determined to high "
                                     "enough precision")
                verbose("alpha=%s" % (alpha), level = 2)
                alpha = alpha.add_bigoh(M)
            except AttributeError:
                pass
        return alpha

    cpdef _richcmp_(_left, _right, int op):
        r"""
        Comparison.

        EXAMPLES:

        Equality of two distributions::

            sage: D = OverconvergentDistributions(0, 5, 10)
            sage: D([1, 2]) == D([1])
            True
            sage: D([1]) == D([1, 2])
            True
            sage: v = D([1+O(5^3),2+O(5^2),3+O(5)])
            sage: w = D([1+O(5^2),2+O(5)])
            sage: v == w
            True
            sage: D = Symk(0,Qp(5,5))
            sage: v = 5 * D([4*5^-1+3+O(5^2)])
            sage: w = D([4+3*5+O(5^2)])
            sage: v == w
            True
        """
        cdef Dist left = _left
        cdef Dist right = _right
        left.normalize()
        right.normalize()
        cdef long rprec = min(left._relprec(), right._relprec())
        cdef long i
        p = left.parent().prime()
        if left.ordp > right.ordp:
            shift = p ** (left.ordp - right.ordp)
            for i in range(rprec):
                lx = shift * left._unscaled_moment(i)
                rx = right._unscaled_moment(i)
                if lx != rx:
                    return richcmp_not_equal(lx, rx, op)
        elif left.ordp < right.ordp:
            shift = p ** (right.ordp - left.ordp)
            for i in range(rprec):
                lx = left._unscaled_moment(i)
                rx = shift * right._unscaled_moment(i)
                if lx != rx:
                    return richcmp_not_equal(lx, rx, op)
        else:
            for i in range(rprec):
                lx = left.moment(i)
                rx = right.moment(i)
                if lx != rx:
                    return richcmp_not_equal(lx, rx, op)
        return rich_to_bool(op, 0)

    def diagonal_valuation(self, p=None):
        """
        Return the largest `m` so that this distribution lies in `Fil^m`.

        INPUT:

        - ``p`` -- (default: ``None``) a positive integral prime

        OUTPUT:

        The largest integer `m` so that `p^m` divides the `0`-th
        moment, `p^{m-1}` divides the first moment, etc.

        EXAMPLES::

            sage: D = OverconvergentDistributions(8, 7, 15)
            sage: v = D([7^(5-i) for i in range(1,5)])
            sage: v
            (O(7^4), O(7^3), O(7^2), O(7))
            sage: v.diagonal_valuation(7)
            4
        """
        if p is None:
            p = self.parent()._p
        n = self.precision_relative()
        return self.ordp + min([n] + [a + self._unscaled_moment(a).valuation(p) for a in range(n)])

    def valuation(self, p=None):
        """
        Return the minimum valuation of any moment.

        INPUT:

        - ``p`` -- (default: ``None``) a positive integral prime

        OUTPUT: integer

        .. WARNING::

            Since only finitely many moments are computed, this valuation may
            be larger than the actual valuation of this distribution.
            Moreover, this valuation may be smaller than the actual
            valuation if all entries are zero to the known precision.

        EXAMPLES::

            sage: D = OverconvergentDistributions(8, 7, 15)
            sage: v = D([7^(5-i) for i in range(1,5)])
            sage: v
            (O(7^4), O(7^3), O(7^2), O(7))
            sage: v.valuation(7)
            4
        """
        if p is None:
            p = self.parent()._p
        n = self.precision_relative()
        if self.parent().is_symk():
            return self.ordp + min([self._unscaled_moment(a).valuation(p) for a in range(n)])
        else:
            return self.ordp + min([n] + [self._unscaled_moment(a).valuation(p) for a in range(n) if not self._unscaled_moment(a).is_zero()])

    def specialize(self, new_base_ring=None):
        """
        Return the image of this overconvergent distribution under
        the canonical projection from distributions of weight `k` to
        `Sym^k`.

        INPUT:

        - ``new_base_ring`` -- (default: ``None``) a ring giving the
          desired base ring of the result

        OUTPUT:

        An element of `Sym^k(K)`, where `K` is the specified base ring.

        EXAMPLES::

            sage: D = OverconvergentDistributions(4, 13)
            sage: d = D([0,2,4,6,8,10,12])
            sage: d.specialize()
            (O(13^7), 2 + O(13^6), 4 + O(13^5), 6 + O(13^4), 8 + O(13^3))
        """
        # self.normalize() # This method should not change self
        k = self.parent()._k
        if k < 0:
            raise ValueError("negative weight")
        if self.precision_absolute() < k + 1:
            raise ValueError("not enough moments")
        V = self.parent().specialize(new_base_ring)
        new_base_ring = V.base_ring()
        if self.precision_relative() == 0:
            return V.zero()
        return V([new_base_ring.coerce(self.moment(j)) for j in range(k + 1)])

    def lift(self, p=None, M=None, new_base_ring=None):
        r"""
        Lift a distribution or element of `Sym^k` to an overconvergent distribution.

        INPUT:

        - ``p`` -- (default: ``None``) a positive integral prime.  If ``None``
          then ``p`` must be available in the parent

        - ``M`` -- (default: ``None``) a positive integer giving the
          desired number of moments. If ``None``, returns a distribution having one
          more moment than this one.

        - ``new_base_ring`` -- (default: ``None``) a ring giving the desired base
          ring of the result. If ``None``, a base ring is chosen automatically.

        OUTPUT:

        - An overconvergent distribution with `M` moments whose image
          under the specialization map is this element.

        EXAMPLES::

            sage: V = Symk(0)
            sage: x = V(1/4)
            sage: y = x.lift(17, 5)
            sage: y
            (13 + 12*17 + 12*17^2 + 12*17^3 + 12*17^4 + O(17^5), O(17^4), O(17^3), O(17^2), O(17))
            sage: y.specialize()._moments == x._moments
            True
        """
        V = self.parent().lift(p, M, new_base_ring)
        k = V._k
        p = V.prime()
        M = V.precision_cap()
        R = V.base_ring()
        moments = [R(self.moment(j)) for j in range(k + 1)]
        zero = R(0)
        moments.extend([zero] * (M - k - 1))
        mu = V(moments)
        # val = mu.valuation()
        # if val < 0:
        #     # This seems unnatural
        #     print("scaling by ", p, "^", -val, " to keep things integral")
        #     mu *= p**(-val)
        return mu

    def _is_malformed(self):
        r"""
        Check that the precision of ``self`` is sensible.

        EXAMPLES::

            sage: D = sage.modular.pollack_stevens.distributions.Symk(2, base=Qp(5))
            sage: v = D([1, 2, 3])
            sage: v._is_malformed()
            False
            sage: v = D([1 + O(5), 2, 3])
            sage: v._is_malformed()
            True
        """
        n = self.precision_absolute()
        for i in range(n):
            if self.moment(i).precision_absolute() < n - i:
                return True
        return False

    def act_right(self, gamma):
        r"""
        The image of this element under the right action by a
        `2 \times 2` matrix.

        INPUT:

        - ``gamma`` -- the matrix by which to act

        OUTPUT:

        - ``self | gamma``

        .. NOTE::

            You may also just use multiplication ``self * gamma``.

        EXAMPLES::

            sage: D = OverconvergentDistributions(4, 7, 10)
            sage: v = D([98,49,21,28,35])
            sage: M = matrix([[1,0], [7,1]])
            sage: v.act_right(M)
            (2*7^2 + 7^3 + 5*7^4 + O(7^5), 3*7^2 + 6*7^3 + O(7^4), 3*7 + 7^2 + O(7^3), 4*7 + O(7^2), O(7))
        """
        return self.parent()._act(self, gamma)

cdef class Dist_vector(Dist):
    r"""
    A distribution is stored as a vector whose `j`-th entry is the `j`-th moment of the distribution.

    The `j`-th entry is stored modulo `p^{N-j}` where `N` is the total number of moments.
    (This is the accuracy that is maintained after acting by `\Gamma_0(p)`.)

    INPUT:

    - ``moments`` -- the list of moments.  If ``check == False`` it
      must be a vector in the appropriate approximation module.

    - ``parent`` -- a :class:`distributions.OverconvergentDistributions_class` or
      :class:`distributions.Symk_class` instance

    - ``ordp`` -- integer;  this *must* be zero in the case of Symk
      of an exact ring

    - ``check`` -- boolean (default: ``True``); whether to validate input

    EXAMPLES::

        sage: D = OverconvergentDistributions(3,5,6) # indirect doctest
        sage: v = D([1,1,1])
    """
    def __init__(self, moments, parent, ordp=0, check=True, normalize=True):
        """
        Initialization.

        TESTS::

            sage: Symk(4)(0)
            (0, 0, 0, 0, 0)
        """
        # if not hasattr(parent,'Element'):
        #     parent, moments = moments, parent

        Dist.__init__(self, parent)
        if check:
            # case 1: input is a distribution already
            if isinstance(moments, Dist):
                ordp = moments._ord_p()
                moments = moments._moments.change_ring(parent.base_ring())
            # case 2: input is a vector, or something with a len
            elif hasattr(moments, '__len__'):
                M = len(moments)
                moments = parent.approx_module(M)(moments)
            # case 3: input is zero
            elif moments == 0:
                moments = parent.approx_module(parent.precision_cap())(moments)
            # case 4: everything else
            else:
                moments = parent.approx_module(1)([moments])
            # TODO: This is not quite right if the input is an inexact zero.
            if ordp != 0 and parent.prime() == 0:
                raise ValueError("cannot specify a valuation shift for an exact ring")

        self._moments = moments
        self.ordp = ordp
        if normalize:
            self.normalize()

    def __reduce__(self):
        r"""
        Used for pickling.

        EXAMPLES::

            sage: D = sage.modular.pollack_stevens.distributions.Symk(2)
            sage: x = D([2,3,4])
            sage: x.__reduce__()
            (<class 'sage.modular.pollack_stevens.dist.Dist_vector'>, ((2, 3, 4), Sym^2 Q^2, 0, False))
        """
        return (self.__class__, (self._moments, self.parent(), self.ordp, False))

    cdef Dist_vector _new_c(self):
        r"""
        Create an empty distribution.

        Note that you MUST fill in the ordp attribute on the resulting distribution.

        OUTPUT:

        - A distribution with no moments.  The moments are then filled
          in by the calling function.

        EXAMPLES::

            sage: D = OverconvergentDistributions(3,5,4) # indirect doctest
            sage: v = D([1,1,1])
        """
        cdef Dist_vector ans = Dist_vector.__new__(Dist_vector)
        ans._parent = self._parent
        return ans

    def _repr_(self):
        r"""
        String representation.

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5]); v
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: repr(v)
            '(1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))'
        """
        valstr = ""
        if self.ordp == 1:
            valstr = "%s * " % (self.parent().prime())
        elif self.ordp != 0:
            valstr = "%s^%s * " % (self.parent().prime(), self.ordp)
        if len(self._moments) == 1:
            return valstr + repr(self._moments[0])
        else:
            return valstr + repr(self._moments)

    def _rational_(self):
        """
        Convert to a rational number.

        EXAMPLES::

            sage: D = Symk(0); d = D(4/3); d
            4/3
            sage: QQ(d)
            4/3

        We get a :exc:`TypeError` if there is more than 1 moment::

            sage: D = Symk(1); d = D([1,2]); d
            (1, 2)
            sage: QQ(d)
            Traceback (most recent call last):
            ...
            TypeError: k must be 0
        """
        if len(self._moments) == 1:
            return QQ(self.moment(0))
        raise TypeError("k must be 0")

    cdef long _relprec(self) noexcept:
        """
        Return the number of moments.

        EXAMPLES::

            sage: D = Symk(4)
            sage: d = D([1,2,3,4,5]); e = D([2,3,4,5,6])
            sage: d == e # indirect doctest
            False
        """
        return len(self._moments)

    cdef _unscaled_moment(self, long n):
        r"""
        Return the `n`-th moment, unscaled by the overall power of `p`
        stored in ``self.ordp``.

        EXAMPLES::

            sage: D = OverconvergentDistributions(4,3,5)
            sage: d = D([3,3,3,3,3])
            sage: d.moment(2) # indirect doctest
            3 + O(3^3)
        """
        return self._moments[n]

    cdef Dist_vector _addsub(self, Dist_vector right, bint negate):
        r"""
        Common code for the sum and the difference of two distributions.

        EXAMPLES::

            sage: D = Symk(2)
            sage: u = D([1,2,3]); v = D([4,5,6])
            sage: u + v # indirect doctest
            (5, 7, 9)
            sage: u - v # indirect doctest
            (-3, -3, -3)
        """
        cdef Dist_vector ans = self._new_c()
        cdef long aprec = min(self.ordp + len(self._moments), right.ordp + len(right._moments))
        ans.ordp = min(self.ordp, right.ordp)
        cdef long rprec = aprec - ans.ordp
        # In the case of symk, rprec will always be k
        V = ans.parent().approx_module(rprec)
        R = V.base_ring()
        smoments = self._moments
        rmoments = right._moments
        # We truncate if the moments are too long; extend by zero if too short
        if smoments.parent() is not V:
            vec = smoments.list(copy=False)[:rprec] + ([R(0)] * (rprec - len(smoments)) if rprec > len(smoments) else [])
            smoments = V(vec)
        if rmoments.parent() is not V:
            vec = rmoments.list(copy=False)[:rprec] + ([R(0)] * (rprec - len(rmoments)) if rprec > len(rmoments) else [])
            rmoments = V(vec)
        # We multiply by the relative power of p
        if self.ordp > right.ordp:
            smoments *= self.parent().prime() ** (self.ordp - right.ordp)
        elif self.ordp < right.ordp:
            rmoments *= self.parent().prime() ** (right.ordp - self.ordp)
        if negate:
            rmoments = -rmoments
        ans._moments = smoments + rmoments
        return ans

    cpdef _add_(self, _right):
        r"""
        Sum of two distributions.

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5]); w = D([3,6,9,12,15])
            sage: v+w
            (4 + O(7^5), 1 + 7 + O(7^4), 5 + 7 + O(7^3), 2 + 2*7 + O(7^2), 6 + O(7))
        """
        return self._addsub(<Dist_vector>_right, False)

    cpdef _sub_(self, _right):
        r"""
        Difference of two distributions.

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5]); w = D([1,1,1,8,8])
            sage: v-w
            (O(7^5), 1 + O(7^4), 2 + O(7^3), 3 + 6*7 + O(7^2), 4 + O(7))
        """
        return self._addsub(<Dist_vector>_right, True)

    cpdef _lmul_(self, Element right):
        r"""
        Scalar product of a distribution with a ring element that coerces into the base ring.

        EXAMPLES::

            sage: D = OverconvergentDistributions(5, 7, 15)
            sage: v = D([1,2,3,4,5]); v
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: 3*v; 7*v
            (3 + O(7^5), 6 + O(7^4), 2 + 7 + O(7^3), 5 + 7 + O(7^2), 1 + O(7))
            7 * (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: v*3; v*7
            (3 + O(7^5), 6 + O(7^4), 2 + 7 + O(7^3), 5 + 7 + O(7^2), 1 + O(7))
            7 * (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
        """
        cdef Dist_vector ans = self._new_c()
        p = self.parent().prime()
        if p == 0:
            ans._moments = self._moments * right
            ans.ordp = self.ordp
        elif right.valuation(p) == Infinity:
            ans._moments = self.parent().approx_module(0)([])
            ans.ordp += self.precision_relative()
        else:
            try:
                v, u = right.val_unit(p)
            except TypeError:  # bug in p-adics: they should accept p here
                v, u = right.val_unit()
            ans._moments = self._moments * u
            ans.ordp = self.ordp + v
            # if the relative precision of u is less than that of self, ans may not be normalized.
        return ans

    def precision_relative(self):
        r"""
        Return the relative precision of this distribution.

        The precision is just the number of moments stored, which is
        also `k+1` in the case of `Sym^k(R)`.  For overconvergent
        distributions, the precision is the integer `m` so that the
        sequence of moments is known modulo `Fil^m`.

        OUTPUT: integer giving the number of moments

        EXAMPLES::

            sage: D = OverconvergentDistributions(2, 11, 15)
            sage: v = D([1,1,10,9,6,15])
            sage: v.precision_relative()
            6
            sage: v = v.reduce_precision(4); v.precision_relative()
            4
            sage: D = Symk(10)
            sage: v = D.random_element()
            sage: v.precision_relative()
            11
        """
        return Integer(len(self._moments))

    def precision_absolute(self):
        r"""
        Return the absolute precision of this distribution.

        The absolute precision is the sum of the relative precision
        (number of moments) and the valuation.

        EXAMPLES::

            sage: D = OverconvergentDistributions(3, 7, base = Qp(7))
            sage: v = D([3,1,10,0])
            sage: v.precision_absolute()
            4
            sage: v *= 7
            sage: v.precision_absolute()
            5
            sage: v = 1/7^10 * v
            sage: v.precision_absolute()
            -5
        """
        return Integer(len(self._moments) + self.ordp)

    cpdef normalize(self, include_zeroth_moment=True):
        r"""
        Normalize by reducing modulo `Fil^N`, where `N` is the number of moments.

        If the parent is Symk, then normalize has no effect.  If the
        parent is a space of distributions, then normalize reduces the
        `i`-th moment modulo `p^{N-i}`.

        OUTPUT: this distribution, after normalizing

        .. WARNING::

            This function modifies the distribution in place as well as returning it.

        EXAMPLES::

            sage: D = OverconvergentDistributions(3,7,10)
            sage: v = D([1,2,3,4,5]) ; v
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: w = v.reduce_precision(3) ; w
            (1 + O(7^5), 2 + O(7^4), 3 + O(7^3))
            sage: w.normalize()
            (1 + O(7^3), 2 + O(7^2), 3 + O(7))
            sage: w
            (1 + O(7^3), 2 + O(7^2), 3 + O(7))
            sage: v.reduce_precision(3).normalize(include_zeroth_moment=False)
            (1 + O(7^5), 2 + O(7^2), 3 + O(7))
        """
        if not self.parent().is_symk() and self._moments != 0:  # non-classical
            if not self._moments:
                return self
            V = self._moments.parent()
            R = V.base_ring()
            n = self.precision_relative()
            p = self.parent()._p
            shift = self.ordp
            if include_zeroth_moment:
                if isinstance(R, pAdicGeneric):
                    self._moments = V([self._moments[i].add_bigoh(n -shift - i) for i in range(n)])
                else:
                    self._moments = V([self._moments[i] % (p ** (n -shift - i)) for i in range(n)])
            else:
                if isinstance(R, pAdicGeneric):
                    self._moments = V([self._moments[0]] + [self._moments[i].add_bigoh(n -shift - i) for i in range(1, n)])  # Don't normalize the zeroth moment
                else:
                    self._moments = V([self._moments[0]] + [self._moments[i] % (p ** (n -shift- i)) for i in range(1, n)])  # Don't normalize the zeroth moment
        return self

    def reduce_precision(self, M):
        r"""
        Only hold on to `M` moments.

        INPUT:

        - ``M`` -- positive integer less than the precision of this
          distribution

        OUTPUT:

        - a new distribution with `M` moments equal to the first `M`
          moments of this distribution.

        EXAMPLES::

            sage: D = OverconvergentDistributions(3,7,10)
            sage: v = D([3,4,5])
            sage: v
            (3 + O(7^3), 4 + O(7^2), 5 + O(7))
            sage: v.reduce_precision(2)
            (3 + O(7^3), 4 + O(7^2))
        """
        assert M <= self.precision_relative(), "not enough moments"

        cdef Dist_vector ans = self._new_c()
        ans._moments = self._moments[:M]
        ans.ordp = self.ordp
        return ans

    def solve_difference_equation(self):
        r"""
        Solve the difference equation. `self = v | \Delta`, where `\Delta = [1, 1; 0, 1] - 1`.

        See Theorem 4.5 and Lemma 4.4 of [PS2011]_.

        OUTPUT:

        - a distribution `v` so that `self = v | Delta` , assuming ``self.moment(0) == 0``.
          Otherwise solves the difference equation for ``self - (self.moment(0),0,...,0)``.

        EXAMPLES::

            sage: D = OverconvergentDistributions(5,7,15)
            sage: v = D(([0,2,3,4,5]))
            sage: g = D._act.actor()(Matrix(ZZ,2,2,[1,1,0,1]))
            sage: w = v.solve_difference_equation()
            sage: v - (w*g - w)
            (O(7^4), O(7^3), O(7^2), O(7))
            sage: v = D(([7,2,3,4,5]))
            sage: w = v.solve_difference_equation()
            sage: v - (w*g - w)
            (7 + O(7^4), O(7^3), O(7^2), O(7))
        """
        # assert self._moments[0][0]==0, "not total measure zero"
        # print("result accurate modulo p^",self.moment(0).valuation(self.p) )
        # v=[0 for j in range(i)]+[binomial(j,i)*bernoulli(j-i) for j in range(i,M)]
        M = self.precision_relative()
        R = self.parent().base_ring()
        K = R.fraction_field()
        V = self._moments.parent()
        v = [K(0) for i in range(M)]
        bern = [bernoulli(i) for i in range(0, M, 2)]
        minhalf = ~K(-2)
        for m in range(1, M):
            scalar = K(self.moment(m) / m)
            # bernoulli(1) = -1/2; the only nonzero odd Bernoulli number
            v[m] += m * minhalf * scalar
            for j in range(m - 1, M, 2):
                v[j] += ZZ(j).binomial(m - 1) * bern[(j - m + 1) // 2] * scalar
        p = self.parent().prime()
        cdef Dist_vector ans
        if p == 0:
            if R in Fields():
                ans = self._new_c()
                ans.ordp = 0
                ans._moments = V(v)
            else:
                newparent = self.parent().change_ring(K)
                ans = newparent(v)
        else:
            ans = self._new_c()
            try:
                ans.ordp = min(a.valuation(p) for a in v)
            except TypeError:
                ans.ordp = 0
            if ans.ordp < 0:
                scalar = K(p) ** (-ans.ordp)
                ans._moments = V([R(a * scalar) for a in v])
            elif ans.ordp > 0:
                scalar = K(p) ** ans.ordp
                ans._moments = V([R(a // scalar) for a in v])
            else:
                ans._moments = V([R(a) for a in v])
            v = ans._moments
            N = len(ans._moments)
            prec_loss = max([N - j - v[j].precision_absolute()
                             for j in range(N)])
            #            print("precision loss = ", prec_loss)
            if prec_loss > 0:
                ans._moments = ans._moments[:(N - prec_loss)]
        return ans


cdef class WeightKAction(Action):
    r"""
    Encode the action of the monoid `\Sigma_0(N)` on the space of distributions.

    INPUT:

    - ``Dk`` -- a space of distributions
    - ``character`` -- data specifying a Dirichlet character to apply to
      the top right corner, and a power of the determinant by which to scale.
      See the documentation of
      :class:`sage.modular.pollack_stevens.distributions.OverconvergentDistributions_factory`
      for more details.
    - ``adjuster`` -- a callable object that turns matrices into 4-tuples
    - ``on_left`` -- whether this action should be on the left
    - ``dettwist`` -- a power of the determinant to twist by
    - ``padic`` -- if ``True``, define an action of `p`-adic matrices (not just integer ones)

    EXAMPLES::

        sage: D = OverconvergentDistributions(4,5,10,base = Qp(5,20)); D
        Space of 5-adic distributions with k=4 action and precision cap 10
        sage: D._act
        Right action by Monoid Sigma0(5) with coefficients in 5-adic Field with capped relative precision 20 on Space of 5-adic distributions with k=4 action and precision cap 10
    """
    def __init__(self, Dk, character, adjuster, on_left, dettwist, padic=False):
        r"""
        Initialization.

        TESTS::

            sage: D = OverconvergentDistributions(4,5,10,base = Qp(5,20)); D # indirect doctest
            Space of 5-adic distributions with k=4 action and precision cap 10
            sage: D = Symk(10) # indirect doctest
        """
        self._k = Dk._k
#        if self._k < 0: raise ValueError("k must not be negative")
        self._adjuster = adjuster
        self._character = character
        self._dettwist = dettwist
        self._p = Dk._p
        self._symk = Dk.is_symk()
        self._actmat = {}
        self._maxprecs = {}
        if character is None:
            self._Np = ZZ(1)  # all of M2Z acts
        else:
            self._Np = character.modulus()
        if not self._symk:
            self._Np = self._Np.lcm(self._p)

        if padic:
            self._Sigma0 = Sigma0(self._Np, base_ring=Dk.base_ring(), adjuster=self._adjuster)
        else:
            self._Sigma0 = Sigma0(self._Np, base_ring=ZZ, adjuster=self._adjuster)
        Action.__init__(self, self._Sigma0, Dk, on_left, operator.mul)

    def clear_cache(self):
        r"""
        Clear the cached matrices which define the action of `U_p`
        (these depend on the desired precision) and the
        dictionary that stores the maximum precisions computed so far.

        EXAMPLES::

            sage: D = OverconvergentDistributions(4,5,4)
            sage: D([1,2,5,3]) * D._act.actor()(Matrix(ZZ,2,2,[1,1,0,1]))
            (1 + O(5^4), 3 + O(5^3), 2*5 + O(5^2), O(5))
            sage: D._act.clear_cache()
        """
        self._actmat = {}
        self._maxprecs = {}

    cpdef acting_matrix(self, g, M):
        r"""
        The matrix defining the action of ``g`` at precision ``M``.

        INPUT:

        - ``g`` -- an instance of
          :class:`sage.matrix.matrix_generic_dense.Matrix_generic_dense`

        - ``M`` -- positive integer giving the precision at which
          ``g`` should act

        OUTPUT:

        - An `M \times M` matrix so that the action of `g` on a
          distribution with `M` moments is given by a vector-matrix
          multiplication.

        .. NOTE::

            This function caches its results.  To clear the cache use
            :meth:`clear_cache`.

        EXAMPLES::

            sage: D = Symk(3)
            sage: v = D([5,2,7,1])
            sage: g = Matrix(ZZ,2,2,[1,3,0,1])
            sage: v * D._act.actor()(g) # indirect doctest
            (5, 17, 64, 253)
        """
        g = g.matrix()
        if g not in self._maxprecs:
            A = self._compute_acting_matrix(g, M)
            self._actmat[g] = {M: A}
            self._maxprecs[g] = M
            return A
        else:
            mats = self._actmat[g]
            if M in mats:
                return mats[M]
            maxprec = self._maxprecs[g]
            if M < maxprec:
                A = mats[maxprec][:M, :M]  # submatrix; might want to reduce precisions
                mats[M] = A
                return A
            if M < 30:  # This should not be hard-coded
                maxprec = max([M, 2 * maxprec])  # This may be wasting memory
            else:
                maxprec = M
            self._maxprecs[g] = maxprec
            mats[maxprec] = self._compute_acting_matrix(g, maxprec)  # could lift from current maxprec
            if M == maxprec:
                return mats[maxprec]
            A = mats[maxprec][:M, :M]  # submatrix; might want to reduce precisions
            mats[M] = A
            return A

    cpdef _compute_acting_matrix(self, g, M):
        r"""
        Compute the matrix defining the action of ``g`` at precision ``M``.

        INPUT:

        - ``g`` -- a `2 \times 2` instance of
          :class:`sage.matrices.matrix_integer_dense.Matrix_integer_dense`

        - ``M`` -- positive integer giving the precision at which
          ``g`` should act

        OUTPUT:

        - ``G`` -- an `M \times M` matrix. If `v `is the vector of moments of a
          distribution `\mu`, then `v*G` is the vector of moments of `\mu|[a,b;c,d]`

        EXAMPLES::

            sage: D = Symk(3)
            sage: v = D([5,2,7,1])
            sage: g = Matrix(ZZ,2,2,[-2,1,-1,0])
            sage: v * D._act.actor()(g) # indirect doctest
            (-107, 35, -12, 5)
        """
        raise NotImplementedError


cdef class WeightKAction_vector(WeightKAction):
    cpdef _compute_acting_matrix(self, g, M):
        r"""
        Compute the matrix defining the action of ``g`` at precision ``M``.

        INPUT:

        - ``g`` -- a `2 \times 2` instance of
          :class:`sage.matrix.matrix_generic_dense.Matrix_generic_dense`

        - ``M`` -- positive integer giving the precision at which
          ``g`` should act

        OUTPUT:

        - ``G`` -- an `M \times M` matrix. If `v` is the vector of moments of a
          distribution `\mu`, then `v*G` is the vector of moments of `\mu|[a,b;c,d]`

        EXAMPLES::

            sage: D = Symk(3)
            sage: v = D([5,2,7,1])
            sage: g = Matrix(ZZ,2,2,[-2,1,-1,0])
            sage: v * D._act.actor()(g) # indirect doctest
            (-107, 35, -12, 5)
        """
        # tim = verbose("Starting")
        a, b, c, d = self._adjuster(g)
        # if g.parent().base_ring().is_exact():
        #     self._check_mat(a, b, c, d)
        k = self._k
        if g.parent().base_ring() is ZZ:
            if self._symk:
                base_ring = QQ
            else:
                base_ring = Zmod(self._p ** M)
        else:
            base_ring = self.underlying_set().base_ring()
        cdef Matrix B = matrix(base_ring, M, M)
        if M == 0:
            return B.change_ring(self.codomain().base_ring())
        R = PowerSeriesRing(base_ring, 'y', default_prec=M)
        y = R.gen()
        # tim = verbose("Checked, made R",tim)
        # special case for small precision, large weight
        scale = (b + d * y) / (a + c * y)
        t = (a + c * y) ** k  # will already have precision M
        cdef long row, col
        # tim = verbose("Made matrix",tim)
        for col in range(M):
            for row in range(M):
                B.set_unsafe(row, col, t[row])
            t *= scale
        # verbose("Finished loop",tim)
        # the changering here is annoying, but otherwise we have to
        # change ring each time we multiply
        B = B.change_ring(self.codomain().base_ring())
        if self._character is not None:
            B *= self._character(a)
        if self._dettwist is not None:
            B *= (a * d - b * c) ** (self._dettwist)
        return B

    cpdef _act_(self, g, _v):
        r"""
        The right action of ``g`` on a distribution.

        INPUT:

        - ``_v`` -- a :class:`Dist_vector` instance, the distribution
          on which to act

        - ``g`` -- a `2 \times 2` instance of
          :class:`sage.matrix.matrix_integer_dense.Matrix_integer_dense`

        OUTPUT: the distribution ``_v * g``

        EXAMPLES::

            sage: D = sage.modular.pollack_stevens.distributions.Symk(2)
            sage: v = D([2,3,4])
            sage: g = Matrix(ZZ,2,2,[3,-1,1,0])
            sage: v * D._act.actor()(g) # indirect doctest
            (40, -9, 2)
        """
        # if g is a matrix it needs to be immutable
        # hashing on arithmetic_subgroup_elements is by str
        if g == 1:
            return _v
        cdef Dist_vector v = <Dist_vector?>_v
        cdef Dist_vector ans = v._new_c()

        try:
            g.set_immutable()
        except AttributeError:
            pass
        v_moments = v._moments
        ans._moments = v_moments * self.acting_matrix(g, len(v_moments))
        ans.ordp = v.ordp
        return ans



# sage_setup: distribution = sagemath-modules
from memory_allocator cimport MemoryAllocator

cdef int *hamming_weights() noexcept

ctypedef unsigned int codeword

cdef struct WordPermutation:
    # A word permutation is a permutation of the bits in a codeword.
    int chunk_num
    int chunk_words
    int degree
    codeword **images
    codeword gate

cdef class BinaryCode:
    cdef MemoryAllocator mem
    cdef codeword *basis
    cdef codeword *words
    cdef int ncols
    cdef int nrows
    cdef int radix
    cdef int nwords

    cdef int is_one(self, int, int) noexcept
    cdef int is_automorphism(self, int *, int *) noexcept
    cpdef int put_in_std_form(self) noexcept
    cdef void _apply_permutation_to_basis(self, object labeling) noexcept
    cdef void _update_words_from_basis(self) noexcept

cdef WordPermutation *create_word_perm(object) noexcept
cdef WordPermutation *create_array_word_perm(int *, int, int) noexcept
cdef WordPermutation *create_id_word_perm(int) noexcept
cdef WordPermutation *create_comp_word_perm(WordPermutation *, WordPermutation *) noexcept
cdef WordPermutation *create_inv_word_perm(WordPermutation *) noexcept
cdef int dealloc_word_perm(WordPermutation *) noexcept
cdef codeword permute_word_by_wp(WordPermutation *, codeword) noexcept
cdef codeword *expand_to_ortho_basis(BinaryCode, int) noexcept

cdef class OrbitPartition:
    cdef MemoryAllocator mem
    cdef int nwords
    cdef int ncols
    cdef int *wd_parent
    cdef int *wd_rank
    cdef int *wd_min_cell_rep
    cdef int *wd_size
    cdef int *col_parent
    cdef int *col_rank
    cdef int *col_min_cell_rep
    cdef int *col_size

    cdef int wd_find(self, int) noexcept
    cdef void wd_union(self, int, int) noexcept
    cdef int col_find(self, int) noexcept
    cdef void col_union(self, int, int) noexcept
    cdef int merge_perm(self, int *, int *) noexcept

cdef class PartitionStack:
    cdef MemoryAllocator mem
    cdef int *wd_ents
    cdef int *wd_lvls
    cdef int *col_ents
    cdef int *col_lvls
    cdef int *basis_locations
    cdef int nwords
    cdef int nrows
    cdef int ncols
    cdef int radix
    cdef int flag
    cdef int *col_degs    #
    cdef int *col_counts  #
    cdef int *col_output  #
    cdef int *wd_degs     #
    cdef int *wd_counts   # These are just for scratch space...
    cdef int *wd_output   #

    cdef int is_discrete(self, int) noexcept
    cdef int num_cells(self, int) noexcept
    cdef int sat_225(self, int) noexcept
    cdef void new_min_cell_reps(self, int, unsigned int *, int) noexcept
    cdef void fixed_vertices(self, int, unsigned int *, unsigned int *, int) noexcept
    cdef int new_first_smallest_nontrivial(self, int, unsigned int *, int) noexcept
    cdef void col_percolate(self, int, int) noexcept
    cdef void wd_percolate(self, int, int) noexcept
    cdef int split_vertex(self, int, int) noexcept
    cdef int col_degree(self, BinaryCode, int, int, int) noexcept
    cdef int wd_degree(self, BinaryCode, int, int, int, int *) noexcept
    cdef int sort_cols(self, int, int) noexcept
    cdef int sort_wds(self, int, int) noexcept
    cdef int refine(self, int, int *, int, BinaryCode, int *) noexcept
    cdef void clear(self, int) noexcept
    cpdef int cmp(self, PartitionStack, BinaryCode) noexcept
    cdef int find_basis(self, int *) noexcept
    cdef void get_permutation(self, PartitionStack, int *, int *) noexcept

cdef class BinaryCodeClassifier:
    cdef MemoryAllocator mem
    cdef int *ham_wts
    cdef int L
    cdef unsigned int *Phi
    cdef unsigned int *Omega
    cdef unsigned int *W
    cdef int radix
    cdef int *Lambda1
    cdef int *Lambda2
    cdef int *Lambda3
    cdef int *w_gamma
    cdef int *c_gamma
    cdef int w_gamma_size
    cdef int *alpha
    cdef int alpha_size
    cdef int *v
    cdef int *e
    cdef int *aut_gp_gens
    cdef int *labeling
    cdef int *base
    cdef int aut_gp_index, aut_gens_size, base_size
    cdef object aut_gp_size

    cdef int Phi_size

    cdef void record_automorphism(self, int *, int) noexcept
    cdef void aut_gp_and_can_label(self, BinaryCode, int) noexcept
